@@ -1410,7 +1410,8 @@ public:
         Mixing of two different non-traditional types is currently prevented.
         This may change in the future.
       */
-      DBUG_ASSERT(item->type_handler()->is_traditional_scalar_type() ||
+      DBUG_ASSERT(item->type_handler()->type_handler_base_or_self()->
+                  is_traditional_scalar_type() ||
                   item->type_handler() == type_handler());
       return true;
     }
@@ -1424,11 +1425,12 @@ public:
                             bool is_eq_func) const override
     {
       // See the DBUG_ASSERT comment in can_optimize_keypart_ref()
-      DBUG_ASSERT(item->type_handler()->is_traditional_scalar_type() ||
+      DBUG_ASSERT(item->type_handler()->type_handler_base_or_self()->
+                  is_traditional_scalar_type() ||
                   item->type_handler() == type_handler());
       return true;
     }
-    void hash(ulong *nr, ulong *nr2)
+    void hash(ulong *nr, ulong *nr2) override
     {
       if (is_null())
         *nr^= (*nr << 1) | 1;
