@@ -1,4 +1,5 @@
 /* Copyright (C) 2006 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
+   Copyright (c) 2020, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -112,13 +113,13 @@ my_bool _ma_fetch_keypage(MARIA_PAGE *page, MARIA_HA *info,
 
   if (lock != PAGECACHE_LOCK_LEFT_UNLOCKED)
   {
-    DBUG_ASSERT(lock == PAGECACHE_LOCK_WRITE || PAGECACHE_LOCK_READ);
+    DBUG_ASSERT(lock == PAGECACHE_LOCK_WRITE || lock == PAGECACHE_LOCK_READ);
     page_link.unlock= (lock == PAGECACHE_LOCK_WRITE ?
                        PAGECACHE_LOCK_WRITE_UNLOCK :
                        PAGECACHE_LOCK_READ_UNLOCK);
     page_link.changed= 0;
     push_dynamic(&info->pinned_pages, (void*) &page_link);
-    page->link_offset= info->pinned_pages.elements-1;
+    page->link_offset= (uint)info->pinned_pages.elements-1;
   }
 
   if (tmp == info->buff)
@@ -153,8 +154,9 @@ my_bool _ma_fetch_keypage(MARIA_PAGE *page, MARIA_HA *info,
     if (page_size < 4 || page_size > share->max_index_block_size ||
         _ma_get_keynr(share, tmp) != keyinfo->key_nr)
     {
-      DBUG_PRINT("error",("page %lu had wrong page length: %u  keynr: %u",
+      DBUG_PRINT("error",("page %lu had wrong page length: %u  page_header: %u  keynr: %u",
                           (ulong) (pos / block_size), page_size,
+                          share->keypage_header,
                           _ma_get_keynr(share, tmp)));
       DBUG_DUMP("page", tmp, page_size);
       info->last_keypage = HA_OFFSET_ERROR;
@@ -438,11 +440,11 @@ my_off_t _ma_new(register MARIA_HA *info, int level,
         Next deleted page's number is in the header of the present page
         (single linked list):
       */
-#ifndef DBUG_OFF
+#ifdef DBUG_ASSERT_EXISTS
       my_off_t key_del_current;
 #endif
       share->key_del_current= mi_sizekorr(buff+share->keypage_header);
-#ifndef DBUG_OFF
+#ifdef DBUG_ASSERT_EXISTS
       key_del_current= share->key_del_current;
       DBUG_ASSERT((key_del_current != 0) &&
                   ((key_del_current == HA_OFFSET_ERROR) ||
@@ -511,8 +513,8 @@ static my_bool _ma_log_compact_keypage(MARIA_PAGE *ma_page,
 
   if (translog_write_record(&lsn, LOGREC_REDO_INDEX,
                             info->trn, info,
-                            log_array[TRANSLOG_INTERNAL_PARTS +
-                                      0].length + extra_length,
+                            (translog_size_t)(log_array[TRANSLOG_INTERNAL_PARTS +
+                                      0].length + extra_length),
                             TRANSLOG_INTERNAL_PARTS + translog_parts,
                             log_array, log_data, NULL))
     DBUG_RETURN(1);
@@ -530,7 +532,7 @@ static my_bool _ma_log_compact_keypage(MARIA_PAGE *ma_page,
    @param min_read_from  Remove all trids from page less than this
 
    @retval 0             Ok
-   retval 1             Error;  my_errno contains the error
+   ®retval 1             Error;  my_errno contains the error
 */
 
 my_bool _ma_compact_keypage(MARIA_PAGE *ma_page, TrID min_read_from)
